/**
 * Module Dependencies
 *
 * @ignore
 */
const moment          = require('moment')
const EventEmitter    = require('structure-event-emitter')
const mixin           = require('structure-mixin')
const r               = require('structure-driver')
const ShortId         = require('structure-short-id-service')
const uuid            = require('structure-uuid')
const Websockets      = require('structure-websocket-server')

function convertDateToDb(d) {

  var date  = (d) ? moment(d) : moment(),
      year  = parseInt(date.format('YYYY')),
      month = parseInt(date.format('M')),
      day   = parseInt(date.format('D'))

  return r.time(year, month, day, process.env.TIME_ZONE)

}

function convertDateToDatabaseTime7(year, month, day, hour = 0, minute = 0, second = 0, timezone = '-04:00') {

  // Allows for the first argument to be "2016-06-30"
  if(arguments.length == 3) {
    var date  = moment(arguments[0])

    year   = parseInt(date.format('YYYY'))
    month  = parseInt(date.format('M'))
    day    = parseInt(date.format('D'))
    hour   = parseInt(arguments[1])
    minute = parseInt(arguments[2])
  }

  // Makes second and timezone optional
  if(arguments.length == 5) {
    year   = parseInt(arguments[0])
    month  = parseInt(arguments[1])
    day    = parseInt(arguments[2])
    hour   = parseInt(arguments[3])
    minute = parseInt(arguments[4])
  }

  year   = parseInt(year)
  month  = parseInt(month)
  day    = parseInt(day)
  hour   = parseInt(hour)
  minute = parseInt(minute)
  second = parseFloat(second)

  return r.time(year, month, day, hour, minute, second, timezone)
}

function convertEpochToDate(e) {
  return moment.unix(e).format('YYYY-MM-DD')
}

function convertEpochToDatabaseTime4(e) {
  var date = this.convertEpochToDate(e)

  return this.convertDateToDatabaseTime4(date)
}

/**
 * RootModel Class
 *
 * @public
 * @class RootModel
 */
class RootModel {

  /**
   * RootModel constructor
   *
   * @public
   * @constructor
   * @param {Object} options - Options
   */
  constructor(options = {}) {

    /**
     * Default parameters
     *
     * @private
     * @type {Object}
     */
    this.defaults = {
      /*
      NOTE: Whether to emit a change feed; if this is false, then web sockets
      will not fire regardless of 'sockets' value. This is because without
      the change feed, the web sockets have nothing to fire.

      In the future sockets may fire without change feeds
      */
      changeFeed: (typeof process.env.CHANGE_FEED == 'boolean') ? process.env.CHANGE_FEED : false,
      sockets: true // whether to emit web socket events if web sockets are available
    }

    /**
     * Assign options
     *
     * @private
     * @type {Object}
     */
    this.options = Object.assign({}, this.defaults, options)

    /**
     * Logs
     *
     * @private
     * @type {String}
     */
    this.logger = options.logger || console
    if(!this.logger.debug) {
      this.logger.debug = console.log.bind(console)
    }

    /**
     * Application ID
     *
     * @private
     * @type {String}
     */
    this.applicationId = options.applicationId

    /**
     * Organization ID
     *
     * @private
     * @type {String}
     */
    this.organizationId = options.organizationId

    /**
     * User ID
     *
     * @private
     * @type {String}
     */
    this.userId = options.userId

    /**
     * Whether the model should fire change feed
     *
     * @private
     * @type {Boolean}
     */
    this.changeFeed = options.changeFeed

    /**
     * List of all cursors opened on all models
     *
     * @private
     * @type {Array}
     */
    this.cursors = options.cursors || this.cursors || []

    /**
     * Keep track of open cursors on model
     *
     * @private
     * @type {Array}
     */
    this._cursors = []

    /**
     * Name of the local key for relations
     *
     * @private
     * @type {String}
     */
    this.localKey = options.localKey || this.localKey

    /**
     * Table relationships
     *
     * @private
     * @type {Boolean}
     */
    this.relations = options.relations

    /**
     * Name of the default table the model works with.
     *
     * @private
     * @type {String}
     */
    this.table = options.table || this.table

    /**
     * Table name
     *
     * @private
     * @type {String}
     */
    if(!this.table) {
      throw new Error('Table must be defined')
    }

    if(this.changeFeed) {
      this.listenToChanges()
    }

  }

  /**
   * Check the existence of a key/value pair, with key being an index
   *
   * @public
   */
  checkExistence(key, val) {

    let exists = false

    return new Promise( async (resolve, reject) => {

      try {

        const filter = {}
        filter[key] = val
        filter.status = 'active'

        const res = await r
          .table(this.table)
          //.getAll(r.args(['active']), {index: 'status'}) <-- Not sure where this index was being set before
          .filter(filter)
          .limit(1)

        if(res.length > 0) exists = true

        resolve({exists})

      }
      catch(e) {
        this.logger.error(e)

        resolve({exists})
      }

    })

  }

  /**
   * Close open cursors
   *
   * @public
   */
  closeCursors() {
    this._cursors.forEach((cursor) => {
      var tokenId = cursor.token

      // Remove this cursor from the full cursor list
      for(let i = 0, l = cursors.length; i < l; i++) {
        var c = cursors[i]

        if(c.token == tokenId) {
          cursors.splice(i, 1)
        }
      }

      cursor.close()
    })

    this._cursors = []
  }

  /**
   * Create, or save, data for the model
   *
   * @public
   * @param {Object} pkg - Data to be saved
   * @param {Object} options - Options
   */
  create(pkg = {}, options = {}) {

    return new Promise( async (resolve, reject) => {

      const tmpPkg = Object.assign({}, pkg)
      const keys   = Object.keys(tmpPkg)

      pkg.createdAt  = r.now()
      pkg.id         = pkg.id || uuid()

      const id = pkg.id

      pkg.sid        = new ShortId().issue(id)
      if(!pkg.status) {
        pkg.status = 'active'
      }
      pkg.updatedAt  = r.now()
      pkg.__version  = process.env.npm_package_version

      const insertOptions = {
        returnChanges: (typeof options.returnChanges == 'boolean') ? options.returnChanges : true
      }

      try {

        /*
        NOTE:
        This will take the package keys, and see if any of the keys belong to a relationship, as defined on the model
        */
        if(this.relations) await this.createRelations(pkg, tmpPkg, keys, id)

        var res = await r.table(options.table || this.table).insert(pkg, insertOptions)
        if(res.errors && res.errors > 0) {
          return reject(res.first_error)
        }

        let doc = {}

        if(insertOptions.returnChanges) {
          doc = res.changes[0].new_val
        }

        resolve(doc)

      }
      catch(err) {
        this.logger.error('Root Model: Could not create', err)

        reject(err)
      }

    })
  }

  createReference(pkg, options = {}) {

    if(!options.table) throw new Error({
      message: 'Reference requires a table'
    })

    return this.create.call(this, pkg, options)

  }

  createRelations(pkg, tmpPkg, keys, localKeyId) {
    //console.error('this', this.table, pkg)
    let refs = []

    for(let i = 0, l = keys.length; i < l; i++) {
      const key = keys[i]

      if(this.relations.belongsTo instanceof Array) {
        for(let j = 0, k = this.relations.belongsTo.length; j < k; j++) {
          const relation = this.relations.belongsTo[j]

          /*
          NOTE:
          If the relationship has a join table,
          has a foreign key, and that foreign key is equal to the pkg key passed in
          */
          if(relation.joinTable && relation.link && relation.link.foreignKey && relation.link.foreignKey == key) {
            const strip = (typeof relation.strip == 'boolean') ? relation.strip : true // Strip join key by default

            // Don't save the join key
            if(strip) {
              delete pkg[key]
            }

            let insert = {}
            insert[relation.link.localKey] = localKeyId // Local Key
            insert[key] = tmpPkg[key] // Foreign Key

            refs.push(r.table(relation.joinTable).insert(insert))
          }
        }
      }

      if(this.relations.hasManyAndBelongsTo instanceof Array) {
        for(let j = 0, k = this.relations.hasManyAndBelongsTo.length; j < k; j++) {
          const relation = this.relations.hasManyAndBelongsTo[j]

          /*
          NOTE:
          If the relationship has a join table,
          has a foreign key, and that foreign key is equal to the pkg key passed in
          */
          if(relation.joinTable && relation.link && relation.link.foreignKey && relation.link.foreignKey == key) {
            const strip = (typeof relation.strip == 'boolean') ? relation.strip : true // Strip join key by default

            // Don't save the join key
            if(strip) {
              delete pkg[key]
            }

            let insert = {}
            insert[relation.link.localKey] = localKeyId // Local Key
            insert[key] = tmpPkg[key] // Foreign Key

            refs.push(r.table(relation.joinTable).insert(insert))
          }
        }
      }

    }

    return Promise.all(refs)

  }

  /**
   * Delete by id
   *
   * @public
   * @param {String} id
   * @return {object} the doc deleted or false if not found
   */
  deleteById(id, options = {}) {

    return new Promise( async (resolve, reject) => {

      try {
        var deletedDoc = await this.updateById(id, {status: 'deleted'})

        if(this.relations) await this.deleteRelations(id)

        return resolve(deletedDoc)
      }
      catch(err) {
        this.logger.error('Root Model: Could not delete', err)

        reject(err)
      }
    })
  }

  /**
   * Delete many by id
   *
   * @public
   * @param {String} id
   * @return {array} the entries deletes
   */
  deleteMany(arrayIds){
    return new Promise( async (resolve, reject) => {
      try {
        var deletedDocs = new Array()
        for(let i=0;i<arrayIds.length;i++){
          deletedDocs.push(this.updateById(arrayIds[i], {status: 'deleted'}))
        }
        return resolve(await Promise.all(deletedDocs))
      }
      catch(err) {
        this.logger.error('Root Model: Could not delete', err)

        reject(err)
      }
    })
  }

  deleteReference(pkg = {}, options = {}) {

    if(!options.table) throw new Error({
      message: 'Reference requires a table'
    })

    return r.table(options.table).filter(pkg).delete()

  }

  /**
   * Delete references
   *
   * @public
   * @param {Object} pkg
   * @param {Array} keys
   * @param {String} localKeyId
   * @return {Array} - References deleted
   */
  deleteRelations(localKeyId) {

    let refs = []

    if(this.relations.belongsTo instanceof Array) {
      for(let j = 0, k = this.relations.belongsTo.length; j < k; j++) {
        const relation = this.relations.belongsTo[j]

        if(relation.joinTable && relation.link && relation.link.foreignKey) {

          let insert = {}
          insert[relation.link.localKey] = localKeyId // Local Key

          refs.push(r.table(relation.joinTable).filter(insert).delete())
        }
      }
    }

    if(this.relations.hasMany instanceof Array) {
      for(let j = 0, k = this.relations.hasMany.length; j < k; j++) {
        const relation = this.relations.hasMany[j]

        if(relation.joinTable && relation.link && relation.link.foreignKey) {

          let insert = {}
          insert[relation.link.localKey] = localKeyId // Local Key

          refs.push(r.table(relation.joinTable).filter(insert).delete())
        }
      }
    }

    if(this.relations.hasManyAndBelongsTo instanceof Array) {
      for(let j = 0, k = this.relations.hasManyAndBelongsTo.length; j < k; j++) {
        const relation = this.relations.hasManyAndBelongsTo[j]

        if(relation.joinTable && relation.link && relation.link.foreignKey) {

          let insert = {}
          insert[relation.link.localKey] = localKeyId // Local Key

          refs.push(r.table(relation.joinTable).filter(insert).delete())
        }
      }
    }

    return Promise.all(refs)

  }

  destroyById(id) {

    return new Promise( async (resolve, reject) => {

      try {
        var deletedDoc = await r.table(this.table).get(id).delete()

        if(this.relations) await this.deleteRelations(id)

        return resolve(deletedDoc)
      }
      catch(err) {
        this.logger.error('Could not delete', err)

        reject(err)
      }

    })

  }

  /*
  NOTE:

  get(id, 'documents')

  get(id, {
    node: 'documents',
    organizationId,
    status: 'draft'
  })

  get('documents', {
    node: 'users'
  })

  get(id, {
    localKey,
    table: 'users'
  })

  // 1:1 Get bucket belonging to organization
  get(bucketId, 'belongsTo', 'organization')
  // 1:Many Organization has many buckets
  get(organizationId, 'hasMany', 'buckets')
  get(organizationId, 'hasMany', {node: 'buckets', ids: [bucket1, bucket2, etc...]})
  get(organizationId, 'hasMany', {node: 'buckets', ids: [bucket1, bucket2, etc...]}, {orderBy: {date: 'desc'}})
  // Many:Many
  get('link', 'users')
  get([group1, group2, etc...], 'link', 'users')
  get([group1, group2, etc...], 'link', {node: 'users', [user1, user2, etc...]})
  */
  get(...args) {

    let query = r.table(this.table)

    switch(args.length) {

      case 1:
        throw new Error({
          message: 'Use getById'
        })

      case 2:
        throw new Error({
          message: 'Many:Many is not supported yet'
        })

      case 3:

        let localKey = null

        // First arguments is an array
        if(args[0] instanceof Array) {
          localKey = r.args(args[0])
        }
        // First argument is an id
        else {
          localKey = args[0]
        }


    }

    return new Promise( async (resolve, reject) => {

      try {
        const res = await query

        resolve(res)
      }
      catch(e) {
        reject(e)
      }

    })

  }

  /**
   * Get all from table
   * Default Limit 100 entries
   * @public
   */
  getAll(ids = [], options = {}) {

    return new Promise( async (resolve, reject) => {

      try {
        var { limit = 0,
              paginate : {offset = 0} = {},
              during : {from, to} = {},
              during = false,
              orderBy = 'updatedAt',
              state = ['active'] } = options,
            topLimit = offset != 0 ? offset + limit : undefined,
            queryLimit = limit !=0 ? limit: 100

        let query = r
          .table(this.table)

        if(ids.length > 0) {

          query = query
            .getAll(r.args(ids))

        }

        query = query
          .orderBy(r.desc(orderBy))

        if(offset && topLimit) {
          query = query
            .slice(offset, topLimit, {right_bound: 'closed'})
        }

        if(queryLimit) {
          query = query
            .limit(queryLimit)
        }

        if(from || to) {
          query = query
            .filter( (row)=> {
              return row('date').during(convertDateToDb.call(this, from), convertDateToDb.call(this, to), { leftBound: 'closed'})
            })
        }

        const res = await query.run()

        resolve(res)

      }
      catch(e) {
        reject(e)
      }

    })

  }

  /**
   * Get model by id
   *
   * @public
   * @param {String} id
   */
  getById(id) {

    if(!id) return Promise.reject({
      message: 'Method requires an Id for lookup'
    })

    // Short ID
    if(id.length <= 10) {

      return new Promise( async (resolve, reject) => {

        try {
          var doc = await r.table(this.table).filter({sid: id}).limit(1)
          return resolve(this.getById(doc[0].id))
        }
        catch(err) {
          this.logger.error(`Could not get object by short ID: ${id}`, err)

          reject({
            message: `Could not get object by short ID: ${id}`
          })
        }

      })
    }

    // Long ID
    else {

      return new Promise( async (resolve, reject) => {

        try {
          let obj = await r.table(this.table).get(id)

          if(!obj) {
            return reject({
              message: `Could not get object by ID: ${id}`
            })
          }

          resolve(obj)
        }
        catch(e) {
          this.logger.error(`Could not get object by ID: ${id}`, e)

          reject(e)
        }

      })

    }

  }

  /**
   * Get reference
   *
   * @private
   * @param {Object} options - Reference filter
   */
  getReference(options = {}) {
    return new Promise( async (resolve, reject) => {

      options.node = options.node || this.name

      try {
        var ref = await r.table('refs').filter(options).limit(1)

        if(ref.length == 0) {
          return reject({
            message: 'Reference not found'
          })
        }

        resolve(ref[0])
      }
      catch(e) {
        this.logger.error('Reference not found', e)

        reject(e)
      }

    })
  }

  /**
   * Get model relationships
   *
   * @private
   * @param {String} type- Relationship type
   */
  getRelations(type) {
    var relations = []

    Object.keys(relations).forEach( (relation) => {

      if(relation == type) {
        relations = relations[relation]
      }

    })

    return relations
  }

  /**
   * Listen to model changes
   *
   * @private
   */
  listenToChanges() {

    r
      .table(this.table)
      .changes()
      .run()
      .then(this.onChangeFeed.bind(this))

  }

  /**
   * When RethinkDB change feed is opened
   *
   * @private
   * @param {Object} cursor - RethinkDB change feed cursor
   */
  onChangeFeed(cursor) {

    cursors.push(cursor)
    this._cursors.push(cursor)

    cursor.each((err, change) => {
      this.emit('change', change.new_val)

      if(this.options.sockets) {
        var wss = new Websockets()
        wss.emit('change', change.new_val)
      }

    })

  }

  /**
   * Create, or save, data for the model
   *
   * @private
   * @param {String} type - Relationship type
   * @param {String} id
   * @param {Object} reference - Reference object
   */
  referenceTo(type, id, reference) {

    return new Promise( (resolve, reject) => {
      var inserts = []

      Object.keys(reference).forEach( async (key) => {
        var referencePkg = {}
        referencePkg[`${this.entityName}Id`] = id // local key
        referencePkg[key] = reference[key]  // foreign key

        switch(type) {
          case 'belongsTo':
            //referencePkg.name = `${key.replace('Id', '')}_has_${this.name}`
            referencePkg.type = 'belongsTo'

            break
        }

        referencePkg.node = this.name

        inserts.push(this.insertReference(referencePkg))
      })

      Promise
        .all(inserts)
        .then(resolve)
        .catch(reject)

    })

  }

  /**
   * Update a model
   *
   * @public
   * @param {String} id
   * @param {Object} pkg - Data to be saved
   * @param {Object} options - Options
   */
  updateById(id, pkg = {}, options = {}) {

    options.returnChanges = (typeof options.returnChanges == 'boolean') ? options.returnChanges : true

    return new Promise( async (resolve, reject) => {

      try {
        // Do not allow these to be updated
        delete pkg.createdAt
        delete pkg.id
        delete pkg.sid

        pkg.updatedAt  = r.now()
        pkg.__version  = process.env.npm_package_version

        const res = await r
          .table(this.table)
          .get(id)
          .update(pkg, options)

        if(!res.changes) {
          this.logger.error('RootModel: RethinkDB was unable to update', res)

          return reject()
        }

        const newVal = res.changes[0].new_val

        /*
        TODO:
        Some modules will receive `newVal` as undefined without this :/
        */
        setTimeout(() => {
          resolve(newVal)
        }, 0)

      }
      catch(e) {
        this.logger.error('Error updating', e)

        reject(e)
      }

    })
  }

  updateMany(data = [],options){

    return new Promise( async (resolve, reject) => {
      try{
        var updatedDocs = []
        for(let i=0;i<data.length;i++){
          updatedDocs.push(this.updateById(data[i].id, data[i].data , options ))
        }

        resolve(await Promise.all(updatedDocs))
      }
      catch(err) {
        this.logger.error('Error updating', err)

        reject(err)
      }
    })
  }

  /**
   * Paginate the results of a query
   * @public
   * @param {Object} query - query object to paginate over
   * @param {Object} options - options for pagination
   * @param {number} options.limit - size of each page, default = 100
   * @param {number} options.page - page number to fetch, default = 1
   * @returns {Promise.Object} - Promise for results of pagination, in the form
   *  {
   *     pages: 3,
   *     current: 1,
   *     previous: false,
   *     next: true,
   *     results: [ ... ]
   *  }
   */
  paginate(query, options = {}) {

    return new Promise( async (resolve, reject) => {

      try {

        const page = options.page || 1
        const limit = options.limit || 100
        const skip = page > 0 ? limit * (page - 1) : 0

        const total = await query.count()

        query = query
          .skip(skip)
          .limit(limit)

        const results = await query.run()

        const totalPages = Math.ceil(total / limit)

        let previous = false
        if (page > totalPages) {
          previous = totalPages
        } else if (page - 1 > 0) {
          previous = page - 1
        }

        resolve({
          pagination: {
            pages: totalPages,
            current: page,
            next: (page + 1 <= totalPages) ? page + 1 : false,
            previous,
            total,
          },
          results
        })

      } catch(e) {

        reject(e)

      }

    })

  }

}

/**
 * Mixin EventEmitter.prototype properties onto RootModel.prototype
 *
 * @ignore
 */
mixin(RootModel, EventEmitter)

module.exports = RootModel
