const Migrations = require('structure-migrations')
const r = require('structure-driver')
const RootModel = require('../../index')

class RootModel2 extends RootModel {}

describe('Integration', function() {

  before(async function() {
    this.migration = new Migrations({
      db: 'test',
      items: {
        tables: [
          {
            action: 'create',
            table: 'root',
            //indexes: ['foo']
          }
        ]
      }
    })

    return this.migration.process()

  })

  beforeEach(function() {

    this.model = new RootModel({
      table: 'root'
    })

  })

  afterEach(function() {
    return this.migration.purge()
  })

  /** @test {RootModel#create} */
  it('should create', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })

    expect(res.foo).to.equal('bar')

  })

  /** @test {RootModel#create} */
  it('should create twice (no duplicate primary key)', async function() {

    let res = await this.model.create({
      foo: 'bar'
    })

    res = await this.model.create({
      foo: 'bar'
    })

    expect(res.foo).to.equal('bar')

  })

  /** @test {RootModel#getById} */
  it('should get by ID', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })

    const res2 = await this.model.getById(res.id)

    expect(res2.foo).to.equal('bar')

  })

  /** @test {RootModel#getAll} */
  it('should get all', async function() {

    await this.model.create({
      foo: 'bar'
    })

    const res = await this.model.getAll()

    expect(res.length > 0).to.be.true

  })

  /** @test {RootModel#update} */
  it('should update by Id', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })

    const res2 = await this.model.updateById(res.id, {
      foo: 'baz',
    })

    const resAll = await this.model.getAll()
    expect(resAll.length === 1).to.be.true // there should be only one record in the Db
    expect(res2.foo).to.equal('baz')

  })

  /** @test {RootModel#update} */
  it('should not allow update sid', async function() {

    const res = await this.model.create({
      foo: 'bar',
    })
    const sid = res.sid

    const res2 = await this.model.updateById(res.id, {
      foo: 'baz',
      sid: 'F3skh5M'
    })

    const resAll = await this.model.getAll()
    expect(resAll.length === 1).to.be.true // there should be only one record in the Db
    expect(res2.sid).to.equal(sid)

  })

  /** @test {RootModel#updateMany} */
  it('should update many', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })

    const res2 = await this.model.create({
      foo: 'baz'
    })

    await this.model.updateMany([
      {id: res.id, data: {foo: 'barbaz'}},
      {id: res2.id, data: {foo: 'bazbar'}}
    ])

    const res3 = await this.model.getById(res.id)
    const res4 = await this.model.getById(res2.id)
    expect(res3.foo).to.equal('barbaz')
    expect(res4.foo).to.equal('bazbar')

  })

  /** @test {RootModel#deleteById} */
  it.skip('should delete by Id', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })
    await this.model.deleteById(res.id)

    const resAll = await this.model.getAll()
    expect(resAll.length === 0).to.be.true

  })

  /** @test {RootModel#deleteMany} */
  it.skip('should delete many', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })

    const res2 = await this.model.create({
      foo: 'bar'
    })

    const res3 = await this.model.create({
      foo: 'bar'
    })

    await this.model.deleteMany([res.id, res2.id, res3.id])

    const resAll = await this.model.getAll()
    expect(resAll.length === 0).to.be.true

  })

  it('should get a reference', async function() {

    await r.table('root').insert({
      applicationId: 2,
      node: 'root',
      organizationId: 1
    })

    const ref = await r.table('root').filter({
      applicationId: 2,
      node: 'root',
      organizationId: 1
    }).limit(1)

    expect(ref).to.be.an('array')
    expect(ref[0].applicationId).to.equal(2)

  })

  it('should be extendable', async function() {

    const model = new RootModel2({
      table: 'root'
    })

    const res = await model.create({
      foo: 'bar'
    })

    const res2 = await model.getById(res.id)

    expect(res2.foo).to.equal('bar')

  })

  it.skip('should not check key/val existence; missing index', async function() {

    await this.model.create({
      fo: 'bar'
    })

    const exists = await this.model.checkExistence('fo', 'bar')

    expect(exists).to.be.false

  })

  it('should check key/val existence', async function() {

    await this.model.create({
      foo: 'bar'
    })

    const result = await Promise
      .all([
        this.model.checkExistence('foo', 'bar'),
        this.model.checkExistence('foo', 'baz')
      ])

    const exists1 = result[0]
    const exists2 = result[1]

    expect(exists1.exists).to.be.true
    expect(exists2.exists).to.be.false

  })

  it('should destroy a record (physical deletion)', async function() {

    const res = await this.model.create({
      foo: 'bar'
    })

    await this.model.destroyById(res.id)

    const resAll = await this.model.getAll()

    expect(resAll.length === 0).to.be.true

  })

  describe('Pagination', function() {

    beforeEach(async function() {

      for (let i=0; i<10; i++) {

        await this.model.create({
          foo: i
        })

      }

    })

    /** @test {RootModel#paginate} */
    it('should paginate over a query with default options', async function() {

      const query = r.table('root')

      const res = await this.model.paginate(query)

      expect(res.results.length).to.equal(10)
      expect(res.pagination.pages).to.equal(1)
      expect(res.pagination.current).to.equal(1)
      expect(res.pagination.total).to.equal(10)
      expect(res.pagination.previous).to.be.false
      expect(res.pagination.next).to.be.false

    })

    /** @test {RootModel#paginate} */
    it('should paginate over a query with specified page/limit', async function() {

      const query = r.table('root')

      const res = await this.model.paginate(query, {page: 1, limit: 3})

      expect(res.results.length).to.equal(3)
      expect(res.pagination.pages).to.equal(4)
      expect(res.pagination.current).to.equal(1)
      expect(res.pagination.total).to.equal(10)
      expect(res.pagination.previous).to.be.false
      expect(res.pagination.next).to.equal(2)

      const res2 = await this.model.paginate(query, {page: 2, limit: 3})

      expect(res2.results.length).to.equal(3)
      expect(res2.pagination.pages).to.equal(4)
      expect(res2.pagination.current).to.equal(2)
      expect(res2.pagination.total).to.equal(10)
      expect(res2.pagination.previous).to.equal(1)
      expect(res2.pagination.next).to.equal(3)

      const res3 = await this.model.paginate(query, {page: 3, limit: 3})

      expect(res3.results.length).to.equal(3)
      expect(res3.pagination.pages).to.equal(4)
      expect(res3.pagination.current).to.equal(3)
      expect(res3.pagination.total).to.equal(10)
      expect(res3.pagination.previous).to.equal(2)
      expect(res3.pagination.next).to.equal(4)

      const res4 = await this.model.paginate(query, {page: 4, limit: 3})

      expect(res4.results.length).to.equal(1)
      expect(res4.pagination.pages).to.equal(4)
      expect(res4.pagination.current).to.equal(4)
      expect(res4.pagination.total).to.equal(10)
      expect(res4.pagination.previous).to.equal(3)
      expect(res4.pagination.next).to.be.false

    })

    /** @test {RootModel#paginate} */
    it('requesting non-existant pages should return []', async function() {

      const query = r.table('root')

      const res = await this.model.paginate(query, {page: 10, limit: 3})

      expect(res.results.length).to.equal(0)
      expect(res.pagination.pages).to.equal(4)
      expect(res.pagination.current).to.equal(10)
      expect(res.pagination.total).to.equal(10)
      expect(res.pagination.previous).to.equal(4)
      expect(res.pagination.next).to.be.false

    })

  })

})
